from .displaced_poles_mapping import displaced_poles_mapping as dp_map
from argparse import ArgumentParser, RawTextHelpFormatter
import numpy as np
import xarray as xr
from datetime import date


def get_dp_grid_ds(lonb, latb, res, NP=None, SP=None, mode='regular'):
    """Produce a displaced poles grid dataset.

    The grid is produced by mapping a grid in the geographical (lon,
    lat) space to the displaced poles space.

    Parameters
    ----------
    lonb: array_like(shape=(2,))
        minimum and maximum longitudes of the geographical grid before
        mapping to the displaced pole space.
    latb: array_like(shape=(2,))
        minimum and maximum latitudes of the geographical grid before
        mapping to the displaced pole space.
    res : float
        Target resolution at the equator before mapping in degrees. The
        resoution will be adapted in both directions so that the `lonb`
        and `latb` values are exactly reached.
    NP : array_like(shape=(2,)), optional
        (lon, lat) coordinates of the displaced north pole in degrees.
        If not given the geographical north pole is used
    SP : array_like(shape=(2,)), optional
        (lon, lat) coordinates of the displaced south pole in degrees.
        If not given, the antipode of NP is used.
    mode : string, optional
        Has to be either 'regular' (default) or 'tele'. 'regular' maps
        a regular (lon, lat) grid to the displaced poles space, 'tele'
        maps a (regular lon, telescopic lat) grid to the displaced
        poles space. The 'telescopic option lets the grid spacing in
        the latitudinal direction vary as the cosine of the latitude,
        exactly like in the longitudinal direction so that the final
        grid is close to isotropic.'

    Returns
    -------
    grid : xarray.Dataset
        Object hosting the generated grid

    """

    # Get the grid mapping class
    # --------------------------
    grd_map = dp_map((0.0, 90.0) if NP is None else NP, SP=SP)

    # Get the grid shape and geographic grid mapping function
    # -------------------------------------------------------
    xi_rho, eta_rho, geo_grd = get_geo_grd(lonb, latb, res, mode)

    # Initiate output grid dataset
    # ----------------------------
    grd_ds = xr.Dataset()
    grd_ds.attrs = {'title': 'ROMS grid generated by get_dp_grid_ds',
                    'date': date.today().strftime('%d-%b-%Y'),
                    'type': 'ROMS grid file',
                    'get_dp_grid_ds: lonb': lonb,
                    'get_dp_grid_ds: latb': latb,
                    'get_dp_grid_ds: res': res,
                    'get_dp_grid_ds: NP': NP,
                    'get_dp_grid_ds: SP': SP,
                    'get_dp_grid_ds: mode': mode}

    # Add grid coordinates at all points and corners
    # ----------------------------------------------
    def add_lon_lat(pt, mode='cell_center'):

        # Compute xi, eta coordinates
        if pt == 'rho':
            xi_0, xi_1 = 1, xi_rho
            eta_0, eta_1 = 1, eta_rho
        elif pt == 'u':
            xi_0, xi_1 = 1.5, xi_rho-0.5
            eta_0, eta_1 = 1, eta_rho
        elif pt == 'v':
            xi_0, xi_1 = 1, xi_rho
            eta_0, eta_1 = 1.5, eta_rho-0.5
        elif pt == 'psi':
            xi_0, xi_1 = 1.5, xi_rho-0.5
            eta_0, eta_1 = 1.5, eta_rho-0.5

        if mode == 'cell_center':
            xi, eta = np.arange(xi_0,xi_1+1), np.arange(eta_0,eta_1+1)
            # Compute geographical lon, lat
            P = np.stack(np.meshgrid(*geo_grd(xi, eta)), axis=-1)
        elif mode == 'cell_corners':
            xi = np.stack((np.arange(xi_0-0.5,xi_1+0.5),
                           np.arange(xi_0+0.5,xi_1+1.5),
                           np.arange(xi_0+0.5,xi_1+1.5),
                           np.arange(xi_0-0.5,xi_1+0.5)), axis=-1)
            eta = np.stack((np.arange(eta_0-0.5,eta_1+0.5),
                            np.arange(eta_0-0.5,eta_1+0.5),
                            np.arange(eta_0+0.5,eta_1+1.5),
                            np.arange(eta_0+0.5,eta_1+1.5)), axis=-1)
            # Compute geographical lon, lat
            n_eta, n_xi, n_crn = eta.shape[0], xi.shape[0], xi.shape[1]
            P = np.empty((n_eta, n_xi, n_crn, 2))
            for k in range(n_crn):
                P[:,:,k,:] = np.stack(np.meshgrid(*geo_grd(xi[:,k], eta[:,k])), axis=-1)
  
        # Compute displaced poles lon, lat
        Q = grd_map.direct_map(P)
        dpp_lon, dpp_lat = Q[...,0], Q[...,1]
        
        # Add to dataset
        PT = pt.upper()
        if mode == 'cell_center':
            lon_name = 'lon_{:s}'.format(pt)
            lat_name = 'lat_{:s}'.format(pt)
            lon_long_name = 'longitude of {:s}-points'.format(PT)
            lat_long_name = 'latitude of {:s}-points'.format(PT)
            dims = ('eta_{:s}'.format(pt), 'xi_{:s}'.format(pt))
        elif mode == 'cell_corners':
            lon_name = 'lon_{:s}_crn'.format(pt)
            lat_name = 'lat_{:s}_crn'.format(pt)
            lon_long_name = 'longitude of {:s}-cells corners clockwise from south-east'.format(PT)
            lat_long_name = 'latitude of {:s}-cells corners clockwise from south-east'.format(PT)
            dims = ('eta_{:s}'.format(pt), 'xi_{:s}'.format(pt), 'crn')
        
        grd_ds[lon_name] = xr.DataArray(dpp_lon, dims=dims, attrs={'long_name': lon_long_name,
                                                                   'units': 'degree_east'})
        grd_ds[lat_name] = xr.DataArray(dpp_lat, dims=dims, attrs={'long_name': lat_long_name,
                                                                   'units': 'degree_north'})
        

    # Add lon, lat at all points
    for pt in 'rho', 'u', 'v', 'psi':
        add_lon_lat(pt, mode='cell_center')
        
    # Add lon, lat corners at rho, u and v-points
    for pt in 'rho', 'u', 'v':
        add_lon_lat(pt, mode='cell_corners')

    # Return grid dataset
    # -------------------
    return grd_ds


def get_geo_grd(lonb, latb, res, mode='regular'):
    """
    Get grid shape and geographical mapping function.

    By grid shape we mean `(xi_rho, eta_rho)` in the ROMS
    nomencalture.  The geographical grid mapping function is the
    function that takes continuous `(xi, eta)` values and outputs the
    geographical `(lon(xi), lat(eta))`, i.e. not yet mapped to the
    displaced pole space.

    Parameters
    ----------
    lonb: array_like(shape=(2,))
        minimum and maximum longitudes of the geographical grid before
        mapping.
    latb: array_like(shape=(2,))
        minimum and maximum latitudes of the geographical grid before
        mapping.
    res : float
        Taget resolution at the equator before mapping in degrees. The
        resoution will be adapted in both directions so that the `lonb`
        and `latb` values are exactly reached.
    mode : string, optional {'regular', 'tele'}, default 'regular'
        Grid spacing in the latitudinal direction. If 'regular', an
        evenly spaced latitudinal grid is assumed, if 'tele', the grid
        spacing varies as the cosine of the latitude (so that the grid
        is isotropic, at least before mapping.)

    Returns
    -------
    xi_1, eta_1 : float
        This values correspond to the xi_rho and eta_rho integers in
        the ROMS nomencalture.
    geo_grd : function
        `geo_grd` takes `xi` and `eta` as parameters and returns a
        `(lon(xi), lat(eta))` tupple.

    Notes
    -----
    `geo_grd: (xi,eta) -> (lon(xi), lat(eta))` satisfies
    `lon(1)=lonb[0]`, `lon(xi_rho)=lonb[1]`, `lat(1)=latb[0]`, and
    `lat(eta_rho)=latb[1]`

    """

    def get_regular_params(x_bnds, res):
        
        delta_x = np.diff(x_bnds)
        nu_1 = np.ceil(delta_x/res) + 1
        d_x = delta_x / (nu_1-1)
        
        return nu_1, x_bnds[0], d_x


    def get_tele_params(x_bnds, res):

        def F(lda):
            return np.arctanh(np.sin(lda))

        theta = np.deg2rad(res)
        x_0, x_1 = np.deg2rad(x_bnds)
        F0, F1 = F(x_0), F(x_1)
        nu_0, nu_1 = 1, 1 + np.ceil((F1-F0)/theta)
        d_x = (F1-F0) / (nu_1-1)
        nu_e = nu_0 - F0/d_x

        return nu_1, nu_e, d_x
        
    
    # regular grid in the lon direction
    xi_1, lon_0, d_lon = get_regular_params(lonb, res)

    if mode == 'regular':
        # regular grid in the lat direction
        eta_1, lat_0, d_lat = get_regular_params(latb, res)

        def geo_grd(xi, eta):
            return (lon_0+d_lon*(xi-1.0), lat_0+d_lat*(eta-1.0))

    elif mode == 'tele':
        # Telescopic grid in the lat direction
        eta_1, eta_e, d_lda = get_tele_params(latb, res)
        
        def geo_grd(xi, eta):
            return (lon_0+d_lon*(xi-1.0), 
                    np.rad2deg(np.arcsin(np.tanh((eta-eta_e)*d_lda))))

    else:
        raise ValueError("kwarg mode has to be either 'regular' or 'tele'")

    return xi_1, eta_1, geo_grd
